
#include <SPI.h>

#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))

volatile long gc;
uint8_t halfsine[201];
uint8_t amp2trinary4_pos[41];
uint8_t amp2trinary4_neg[41];
uint8_t trit[3] = {4,0,1};        // not compliment logic, so -1 = B10, 0 = B00, +1 = B01

int vv;                      // global for DAC output

uint16_t pcount;
long pcount_acc;
int pcount_delta, phase_error, old_phase_error;
int pcint,new_delta, p_lock;

uint8_t v1low,int_f,uf;
uint16_t vpwr;
bool working;
float dc_o,dc_s,dcvs;

void setup() 
  {  
    int i,j,k,m,t1,t2,t3,t4;
    float t, u;
    char s[10];

  dcvs = 0.0;     // for now, these values will do. = 0.0 for no effect
  dc_o = 0.0;     // calibration for DC supply voltage sample, input into ADC ch5 
  dc_s = 0.0;     //
    
  ADCSRA =  bit (ADEN);   // turn ADC on, freerunning
  ADCSRA |= bit (ADPS2) |  bit (ADPS1) | bit (ADPS0);  // Prescaler of 128

  pinMode(9,OUTPUT) ;  // LDAC pin 8
  // D10, CS, pin3 on DAC
  // D11, data out, pin 5 on DAC
  // D13, SCK, to pin 4 on DAC
  // The SPI is for the DAC chip, used to make nice looking DSO trace of output sine wave
  SPI.begin() ;                        
  SPI.setBitOrder(MSBFIRST) ;          
  SPI.setClockDivider(SPI_CLOCK_DIV2) ;
  SPI.setDataMode(SPI_MODE0) ;         
  TIMSK0 = 0;
  pinMode(8,OUTPUT);      // PORTB,0
  pinMode(2,INPUT);       // for mains sync interrupt
    
  DDRD = 0xf0;            // all 8 bits to be output. 4 bits out top 4 of PORTD
  DDRC = 0x0f;            // and 4 bits out PORTC, the usually configured ADC pins
  
  TCCR1A =  _BV(WGM11);   // a nice 20kHz timer, which calls the below interrupt
  TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS10);
  ICR1 = 800;
  TIMSK1 |= (1 << TOIE1); // enable the interrupt

  EICRA = 3;               // interrupt on falling edge from pin D2
  EIMSK |= bit (INT0);     // enable it  

  for(t=0.0,i=0; i <= 200; i++,t +=   3.14159/ (float) (201))
    {   
    u = 40.0 * sin(t);   // 1/2 wave sine lookup table, scaled to (8 bits - 1)
    halfsine[i] =  (int)u;  
    }
 
  pcount_acc = 0;      // fractional integer math. Each PWM, pcount_delta is added to pcount_acc and divided by 1024
  pcount_delta = 1024; // in effect we are getting an addition of 1. But pcount_delta can be modified, a little bigger or smaller..
  pcount = 0;
  vv = 0;
  uf = 0;
  vpwr = 200;           // for now. It will be changed depending on DC supply voltage
  p_lock = 0;          // phase lock flag = 0 = NOT locked
  old_phase_error = 0; // used in PID, to obtain derivative

  // set up the 4 inverter bridge control signalling
  // will use this later to determine the signalling for a selected output voltage
  
  amp2trinary4_pos[0] = 0; 
  amp2trinary4_neg[0] = 0; 
  t2=t3=t4=0;
  t1 = 1;
  for(i=1; i <= 40; i++)
    {
    k = 0;  
    m = 0;
    if (t1 ==  1) k += B00000001;
    if (t1 == -1) k += B00000010;
    if (t2 ==  1) k += B00000100;
    if (t2 == -1) k += B00001000;
    if (t3 ==  1) k += B00010000;
    if (t3 == -1) k += B00100000;
    if (t4 ==  1) k += B01000000;
    if (t4 == -1) k += B10000000;

    if (t1 == -1) m += B00000001;
    if (t1 ==  1) m += B00000010;
    if (t2 == -1) m += B00000100;
    if (t2 ==  1) m += B00001000;
    if (t3 == -1) m += B00010000;
    if (t3 ==  1) m += B00100000;
    if (t4 == -1) m += B01000000;
    if (t4 ==  1) m += B10000000;
    amp2trinary4_pos[i] = k; 
    amp2trinary4_neg[i] = m; 
    t1++; 
    if (t1 > 1) 
      {
        t1 = -1;
        t2++;
        if (t2 > 1)
          {
            t2 = -1;
            t3++; 
            if (t3 > 1)
              {
                t3 = -1;
                t4++;
              }
          }      
      }   
    }
  }

// print 8 bits, all aligned nicely.
void check_bits()
  {
  int i,j;    
  for (i = 0; i <= 40; i++)
    {
      Serial.print(i);
      Serial.write('\t');
      for (j=7; j >=0; j--)
        {
          if ((amp2trinary4_pos[i] & (1 << j)) > 0)
            Serial.print("1");
          else
            Serial.print("0");
        }
      Serial.write('\t');
      for (j=7; j >=0; j--)
        {
          if ((amp2trinary4_neg[i] & (1 << j)) > 0)
            Serial.print("1");
          else
            Serial.print("0");
        }
      Serial.println();
    }
  }

// 20kHz basic clock
// 1/400th = 50 Hz
ISR(TIMER1_OVF_vect)        
  { 
  uint8_t b;
  int a;

  pcount_acc += pcount_delta; // count through 1/2 wave sine table, 200 entries. At 20KHz, = 50Hz output 
  pcount = (pcount_acc >> 10); 

  if(pcount >= 200)          // pcount  will = NPWM, at each 1/2 wave, at the start
    {
    pcount = 0;               // reset counter
    pcount_acc = 0;           // and the fractional math part of the counter
    if (v1low == 1)           // if first 1/2 wave..
      {
        //sbi(PORTB,0);
        //cbi(PORTB,0);
        v1low = 0;             // toggle to 2nd half wave
        int_f = 1;               // this is here, last to execute, to ensure this PWM code runs first and is not delayed by
        uf=1;
       }                         // subsequent AC sync interrupt processing. 
    else
      {
       v1low = 1;             // toggle to 1st half wave
      }
    }
  // get sin wave amplitude from halfsine lookup table
  // which is scaled from 0 - 40 - 0, and this suits the gate drive lookup table with 40 entries. 
  // scale by vpwr, 0 - 255 possible settings, then divide by 256
  // vpwr comes from DC supply voltage deviation from some standard setting, calculated in loop()
  // loop() is executed as often as possible, outside both interrupt blocks
  //
  a = (halfsine[pcount] * vpwr) >> 8;
  // ..
  // now get the 4 x full bridge drive equivalent, a la Warpspeed's  concept.
  // binary = 0,1   trinary = -1,0,1
  if (v1low == 0)
    {
      vv = 2048 + 50*a;
      b = amp2trinary4_pos[a];
    }
  else
    {
      vv = 2048 - 50*a;
      b = amp2trinary4_neg[a];
    }
  // 4 bits out one port, 4 bits out the other  
  // testing code. Production code will run on an Arduino Mega with enough digital pins on one port
  // and int1 available for mains sync. If no mains sync, it will run on the Uno or Nano.
  // I split this into 2 ports because I need pin D2 for int1 signal input used in mains sync.
  // if no mains sync then the below 2 lines would be
  // PORTD = b;
  // and work perfectly
  //
  PORTD = b & 0xf0;    
  PORTC = b & 0x0f;
  }



ISR (INT0_vect)
  {
  if  (int_f == 0) 
    {  
    return;  // only once per 50 Hz cycle. Noise! much noise...
    }
 
  int_f = 0;
  pcint = pcount +  (v1low == 1?200:0) - 200;   // get position in 50Hz waveform, -200 to 200 counts
                                                // pcint is position in output waveform when this interrupt occurs
  phase_error = (pcint - 4);    // subtract setpoint, 0 for my setup, in counts
                                // mains sync sample will have some low pass filter in effect, this gives a phase delay.
                                // setpoint = 0, will be different, effected by RC LP filter delay at least, other influences?              
  new_delta =   (phase_error/4 + (phase_error - old_phase_error)); // PID, no I. P = 0.25, D = 1.0.  
  //the divide by 4 is compiled as 2 x looped arithmetic shifts
   
  if (new_delta > 80) 
    new_delta = 80;       // Keeps things safe. When 80 or higher is used, my inverter spasms. Safe thanks to current limited power supply
  if (new_delta < -80) 
    new_delta = -80;     // even this small a range of changes still lets it sync to mains is less than 1/2 second.    

  pcount_delta = pcount_delta - new_delta;  // apply PID output to the process, i.e. modify AC output frequency 
  if(pcount_delta < 900) pcount_delta = 900;  // clamp change to output frequency to reasonable limits
  if(pcount_delta > 1100) pcount_delta = 1100;  // this range is approx 40 - 58 Hz, with 1024 = 50Hz mid point
  old_phase_error = phase_error;
  if (abs(phase_error) < 5 && abs(new_delta) < 5)
    {
      p_lock = 1;
      //sbi(PORTD,4);              // tell world I think I'm in sync. drive relay, light LED
    }
  else
    {
      p_lock=0;
      //cbi(PORTD,4);             // need to chose a free output pin..
    }
  }  

  
void loop() 
  {
  float ch0;
  float dcv,pwr;
    
   if (!working)                // async ADC conversion, non blocking
    {
    ADMUX =   bit (REFS0) | 5;  // chose ADC ch5       
    bitSet (ADCSRA, ADSC);      // start an ADC conversion
    working = true;
    }
    
  dac_a(vv);    // output a sinewave, via the MCP4912 DAC chip, just for fun and education. Not needed for inverter function
  
  if (uf == 1 && bit_is_clear(ADCSRA, ADSC) )  // only execute this once each 50Hz wave form and when ch5 ADC is ready
    {
      uf=0;
      working = false;    
      ch0 = (float)ADC / 1024.0;        // scale 0 - 5V to 0.0 to 1.0
      
      // This is a 1,3,9,27 ratio tranformer build. Total voltage out is 40 units.
      // we want 240V, so it's 6V/unit at full power
      // Good inverter design has some primary voltage headroom, maybe 20%
      // so let's set normal low voltage cutoff at 48.5V, this will be our 80% power setting
      // 80% of 40 units = 32 units.
      // 48.5V primary = 34.3V AC RMS primary voltage.
      // 240V needed, can only drive it to 32 units. This means 240/32 V per unit = 7.5 V/unit
      // the transformer secondary output voltage will be
      // 7.5 x 1, 7.5 x 3, 7.5 x 9, 7.5 x 27
      // 7.5V AC rms, 22.5V AC, 67.5V AC, 202.5V AC
      // All secondary windings need to be capable of the intended AC output current.
      // Wind the primaries to suit the above output voltages...
      
      dcv = ch0 * dc_s + dc_o;    // calibrate dc sample
      
      // now we have the DC volts.
      // obtain a power setting, where 0.8 = 48.5V, the setting will be bigger if dcv < 48.5
      // and smaller when > 48.5. The scale factor of these adjustments to be determined.
      
      pwr = 0.8 - (48.5 - dcv)*dcvs;   
      
      if (pwr < 0.2) pwr = 0.2;       // prevent it going too low for some reason
      if (pwr > 0.99) pwr = 0.99;     // and too high
      
      vpwr = int(256.0 * pwr);        // scale and convert to int, ready for fast maths in the 20kHz interrupt  
    }
  }

void dac_a(int i)
      {
      sbi(PORTB,1); // output voltage via MCP4912 DAC chip's A channel
      cbi(PORTB,2);
      SPDR = (i >> 8)|0x30;
      asm volatile("nop"); // See transfer(uint8_t) function
      while (!(SPSR & _BV(SPIF))) ; 
      SPDR = i & 0xff;
      asm volatile("nop");
      while (!(SPSR & _BV(SPIF))) ;
      sbi(PORTB,2);
      cbi(PORTB,1);      
     }
